import asyncio
import datetime as dtm
from typing import Optional


class WallClock:

    def __init__(self, live_mode: bool = True, start_time: Optional[dtm.datetime] = None, end_time: Optional[dtm.datetime] = None, tzinfo: Optional[dtm.tzinfo] = None):
        self.live_mode = live_mode
        self.tzinfo = tzinfo

        if not live_mode:
            self.current_time = start_time
            self.end_time = end_time if end_time is not None else dtm.datetime.now()

            if self.tzinfo is not None:
                self.current_time = self.current_time.astimezone(self.tzinfo)
                self.end_time = self.end_time.astimezone(self.tzinfo)

        self.on_time_change = None
        self.on_before_time_change = None

    def now(self) -> dtm.datetime:
        if self.live_mode:
            return dtm.datetime.now(self.tzinfo)
        else:
            if self.current_time > dtm.datetime.now(self.tzinfo):
                raise ValueError("backtesting time is must be less than current time")

            return self.current_time

    def today(self) -> dtm.date:
        return self.now().date()

    def is_alive(self) -> bool:
        if self.live_mode:
            return True
        else:
            return self.current_time <= self.end_time

    async def sleep(self, seconds: float):
        if self.live_mode:
            await asyncio.sleep(seconds)
        else:
            before = self.current_time
            after = before + dtm.timedelta(seconds=seconds)

            if self.on_before_time_change:
                self.on_before_time_change(before, after)

            self.current_time = after

            if self.on_time_change:
                self.on_time_change(self.current_time, before)



from dataclasses import dataclass
from enum import Enum
from decimal import Decimal
from typing import Optional
import datetime as dtm


class OrderSide(Enum):
    SELL = 1
    """ 매도 """
    BUY = 2
    """ 매수 """


class OrderType(Enum):
    LIMIT = 0
    """ 지정가 """
    LIMIT_IOC = 11
    """ 지정가, 즉시(일부)체결 그리고 잔량취소 """
    LIMIT_FOK = 12
    """ 지정가, 즉시 전량 체결 또는 전량취소 """
    LIMIT_CONDITIONAL = 2
    """ 조건부지정가, 지정가 주문을 넣고 장 마감까지 체결되지 않은 경우 동시호가 가격으로 체결 시킴 """
    MARKET = 1
    """ 시장가 """
    MARKET_IOC = 13
    """ 시장가, 즉시 (일부)체결 그리고 잔량취소 """
    MARKET_FOK = 14
    """ 시장가, 즉시 전량체결 또는 전량취소 """
    PRIMARY_PRICE = 3
    """ 최우선지정가, 반대방향의 1차호가로 주문 제출 """
    BEST_PRICE = 4
    """ 최유리지정가, 같은방향의 1차호가로 주문 제출 """
    BEST_PRICE_IOC = 15
    """ 최유리지정가, 즉시 (일부)체결 그리고 잔량취소"""
    BEST_PRICE_FOK = 16
    """ 최유리지정가, 즉시 전량체결 또는 전량취소"""
    PRE_MARKET = 5
    """ 장전단일가, 장전 단일가 주문 """
    AFTER_MARKET = 6
    """ 시간외종가, 시간외 종가 주문 """
    AFTER_MARKET_SINGLE_PRICE = 7
    """ 시간외단일가, 시간외 단일가 주문 """
    SELF_STOCK = 8
    """ 자사주, 자사주 매수 또는 매도 주문 """
    SELF_STOCK_S_OPTION = 9
    """ 자사주 S옵션, 자사주 매수 또는 매도 주문 """
    SELF_STOCK_MONETARY_TRUST = 10
    """ 자기주식금융신탁 """
    MOO = 31
    """ Market On Open, 해외 주식 주문 유형 (매도시에만 가능)"""
    LOO = 32
    """ Limit On Open, 해외 주식 주문 유형 """
    MOC = 33
    """ Market On Close, 해외 주식 주문 유형 (매도시에만 가능) """
    LOC = 34
    """ Limit On Close, 해외 주식 주문 유형 """


class OrderRequestType(Enum):
    NEW = 1
    """ 신규주문 """
    MODIFY = 2
    """ 주문수정 """
    CANCEL = 3
    """ 주문취소 """


class TransactionStatus(Enum):
    ORDER = 1
    """ 주문 """
    EXECUTION = 2
    """ 체결 """


@dataclass
class StockPosition:
    """
    주식 보유 종목 정보를 담는 데이터 클래스입니다.
    """

    asset_code: str
    """ 종목 코드 """
    asset_name: str
    """ 종목 이름 """
    quantity: int
    """ 보유 수량 """
    sell_possible_quantity: int
    """ 매도 가능 수량 """
    average_purchase_price: Decimal
    """ 평균 매입 가격 """
    current_price: Optional[int | Decimal] = None
    """ 현재 가격 """
    current_value: Optional[int | Decimal] = None
    """ 현재 가치 """
    current_pnl: Optional[Decimal] = None
    """ 현재 손익률 """
    current_pnl_value: Optional[int | Decimal] = None
    """ 현재 손익 금액 """


@dataclass(kw_only=True)
class OverseasStockPosition(StockPosition):
    """
    해외 주식 보유 종목 정보를 담는 데이터 클래스입니다.
    """

    exchange: str
    """ 거래소 """
    currency: str
    """ 통화 """


@dataclass
class StockOrder:
    """
    주식 주문 정보를 담는 데이터 클래스입니다.
    """

    order_no: str
    """ 주문 번호 """
    asset_code: str
    """ 종목 코드 """
    side: OrderSide
    """ 주문 방향 """
    price: int | Decimal
    """ 주문 가격 """
    quantity: int
    """ 주문 수량 """
    filled_quantity: int
    """ 체결 수량 """
    pending_quantity: int
    """ 미체결 수량 """
    order_time: dtm.datetime
    """ 주문 시각 """
    filled_price: Optional[int | Decimal] = None
    """ 체결 가격 """
    current_price: Optional[int | Decimal] = None
    """ 현재 가격 """
    is_pending: Optional[bool] = None
    """ 미체결 여부 """
    org_order_no: Optional[str] = None
    """ 원주문번호 """
    order_type: Optional[OrderType] = None
    """ 주문 유형 """
    req_type: Optional[OrderRequestType] = None
    """ 주문 요청 유형 """


@dataclass(kw_only=True)
class OverseasStockOrder(StockOrder):
    """
    해외 주식 주문 정보를 담는 데이터 클래스입니다.
    """

    exchange: str
    """ 거래소 """
    currency: str
    """ 통화 """
    reject_reason: Optional[str] = None
    """ 거부 사유 """


@dataclass
class OrderEvent:
    asset_code: str
    """ 종목 코드 """
    order_no: str
    """ 주문 번호 """
    side: OrderSide
    """ 매도 / 매수 """
    order_type: OrderType
    """ 주문 유형 """
    quantity: int
    """ 주문 수량 """
    price: int
    """ 주문 가격 """
    event_type: str
    """ 이벤트 유형 - accepted, executed, cancelled, rejected """
    account_no: Optional[str] = None
    """ 계좌 번호 """
    filled_quantity: Optional[int] = None
    """ 체결 수량 """
    filled_price: Optional[int] = None
    """ 체결 가격 """
    filled_time: Optional[dtm.datetime] = None
    """ 체결 시각 """
    org_order_no: Optional[str] = None
    """ 원주문번호 """
    average_purchase_price: Optional[Decimal] = None
    """ 평균 매입 가격 """

    @staticmethod
    def with_pending_order(account_no: str, order: StockOrder):
        """
        미체결 주문 정보로 이벤트 생성

        Args:
            account_no (str): 계좌 번호
            order (StockOrder): 주문 정보
        """
        return OrderEvent(
            asset_code=order.asset_code,
            order_no=order.order_no,
            side=order.side,
            order_type=order.order_type,
            quantity=order.quantity,
            price=order.price,
            status="accepted",
            account_no=account_no,
            filled_quantity=order.filled_quantity,
            filled_price=order.filled_price,
            filled_time=None,
            org_order_no=order.org_order_no,
        )


@dataclass
class TradingHistory:
    """거래내역 정보를 담는 데이터 클래스입니다"""

    date: str
    """ 거래일자 (YYYYMMDD) """
    order_no: str
    """ 주문번호 """
    side: str
    """ 매도:sell / 매수:buy """
    asset_code: str
    """ 종목코드 """
    quantity: int
    """ 수량 """
    filled_price: int
    """ 체결가격 """
    average_purchase_price: Optional[float] = None
    """ 평균 매입가 """
    tax: Optional[float] = None
    """ 제세금 (증권거래세+지방소득세) """
    fee: Optional[float] = None
    """ 수수료 """
    pnl: Optional[float] = None
    """ 손익금액 """
    pnl_rate: Optional[float] = None
    """ 손익률 """
    executed_time: Optional[int] = None
    """ 체결시간 """
    partial: Optional[bool] = None
    """ 일부체결 여부 """



from dataclasses import asdict
from decimal import ROUND_HALF_UP, Decimal
from pyqqq.brokerage.ebest.simple import EBestSimpleDomesticStock
from pyqqq.brokerage.kis.simple import KISSimpleDomesticStock
from pyqqq.data.domestic import get_tickers
from pyqqq.datatypes import *
from pyqqq.utils.api_client import raise_for_status, send_request
from pyqqq.utils.array import find
from pyqqq.utils.logger import get_logger, logging
from pyqqq.utils.market_schedule import get_market_schedule
from typing import Dict, Optional, List
import asyncio
import os
import pyqqq.config as c


class TradingTracker:
    """
    거래 내역 추적을 위한 클래스입니다

    주문 이벤트를 수신하여 보유 포지션과 미체결 주문을 관리하고 거래 내역을 기록합니다.

    Args:
        simple_api (EBestSimpleDomesticStock | KISSimpleDomesticStock): 간편 거래 API 객체
        fee_rate (Decimal): 증권사 수수료율 (기본값: 0.015%)
    """

    logger = get_logger("trading_tracker", level=logging.INFO)

    def __init__(
        self,
        simple_api: EBestSimpleDomesticStock | KISSimpleDomesticStock,
        fee_rate: Decimal = Decimal(0.00015),  # 뱅키스, LS증권 수수료율 0.015%
    ):
        self.positions: List[StockPosition] = []
        """ 보유 포지션 목록 """

        self.pending_orders: List[StockOrder] = []
        """ 미체결 주문 목록 """

        self.on_pending_order_update: Optional[callable] = None
        """ 미체결 주문 업데이트 이벤트 callback

        Args:
            status (str): 이벤트 상태. 'accepted', 'cancelled', 'completed', 'partial'
            order (StockOrder): 주문 정보
        """

        self.on_position_update: Optional[callable] = None
        """ 포지션 업데이트 이벤트 callback

        Args:
            type (str): 이벤트 타입. 'added', 'modified', 'removed'
            position (StockPosition): 포지션 정보
        """

        self.task: asyncio.Task = None
        """ 백그라운드로 실행되는 거래 이벤트 모니터링 Task """

        self.simple_api = simple_api
        self.stop_event = asyncio.Event()
        self.account_no = None
        self.fee_rate = fee_rate  # 증권사 수수료율
        self.tax_rate = Decimal("0.0018")  # KOSPI, KOSDAQ 매도시 거래세율 0.18%
        self.tickers: Dict[str, Dict] = {}  # 종목 코드별 종목 정보
        self.ticker_date: dtm.datetime = None  # 종목 정보 갱신 시간

    async def start(self):
        """
        거래 내역 추적을 시작합니다
        """
        if isinstance(self.simple_api, EBestSimpleDomesticStock):
            account_info = self.simple_api.get_account()
            self.account_no = account_info["account_no"]
        elif isinstance(self.simple_api, KISSimpleDomesticStock):
            self.account_no = (
                self.simple_api.account_no + self.simple_api.account_product_code
            )

        self.logger.info(f"Trading tarcker started! Account No: {self.account_no}")

        self._fetch_tickers()
        self._sync_positions_and_pending_orders()

        if len(self.pending_orders) > 0:
            self.logger.info("Initial pending orders:")
            for o in self.pending_orders:
                self.logger.info(
                    f"- {o.order_no}({o.org_order_no})\t{o.side}\t{o.asset_code}\t{o.filled_quantity}/{o.quantity}\t{o.is_pending}"
                )

        self.tasks = [
            asyncio.create_task(self._monitor_trading()),
            asyncio.create_task(self._monitor_schedule()),
        ]

    def _fetch_tickers(self):
        df = get_tickers(dtm.date.today())
        df.reset_index(inplace=True)
        for d in df.to_dict(orient="records"):
            self.tickers[d["code"]] = d

        self.ticker_date = dtm.datetime.now()

    def _sync_positions_and_pending_orders(self):
        self.positions = self.simple_api.get_positions()
        for p in self.positions:
            p.current_price = None
            p.current_value = None
            p.current_pnl = None
            p.current_pnl_value = None

        self.pending_orders = self.simple_api.get_pending_orders()

    async def stop(self):
        """
        거래 내역 추적을 중지합니다
        """
        self.stop_event.set()
        for t in self.tasks:
            t.cancel()

        await asyncio.gather(*self.tasks)

    async def _monitor_trading(self):
        try:
            async for event in self.simple_api.listen_order_event(self.stop_event):
                self._handle_order_event(event)
        except asyncio.CancelledError:
            return
        except Exception as e:
            self.logger.exception(f"Error on handling order event: {e}")

    async def _monitor_schedule(self):
        """거래 시간대별 작업을 위한 스케줄을 모니터링합니다"""
        while not self.stop_event.is_set():
            market_schedule = get_market_schedule(dtm.date.today())

            if not market_schedule.full_day_closed:
                # 정규장 시작 30분 전 종목정보 갱신
                ticker_fresh_time = (
                    dtm.datetime.combine(dtm.date.today(), market_schedule.open_time)
                    - dtm.timedelta(minutes=30)
                ).time()

                ticker_refresh_seq_no = self._calc_clock_seq(t=ticker_fresh_time)

                # 정규장 종료 시 주문, 포지션 정보 갱신
                market_close_sync_seq_no = self._calc_clock_seq(
                    t=market_schedule.close_time
                )

                # 시간외 거래 종료 시 주문, 포지션 정보 갱신
                after_market_close_sync_seq_no = self._calc_clock_seq(
                    t=dtm.time(18, 0, 0)
                )

            try:
                seq_no = self._calc_clock_seq()

                if market_schedule.full_day_closed:
                    pass

                elif seq_no == ticker_refresh_seq_no:
                    self._fetch_tickers()
                    self._sync_positions_and_pending_orders()

                elif seq_no == market_close_sync_seq_no:
                    self._sync_positions_and_pending_orders()
                    self.save_positions()

                elif seq_no == after_market_close_sync_seq_no:
                    self._sync_positions_and_pending_orders()

                await asyncio.sleep(60)
            except asyncio.CancelledError:
                return
            except Exception as e:
                self.logger.exception(f"Error on monitoring schedule: {e}")

    def _calc_clock_seq(
        self,
        interval=dtm.timedelta(seconds=60),
        t: dtm.time = None,
    ):
        """현재 시각을 기준으로 시간대별 시퀀스 번호를 계산합니다"""
        midnight = dtm.datetime.combine(dtm.date.today(), dtm.time.min)
        clock_time = (
            dtm.datetime.now()
            if t is None
            else dtm.datetime.combine(dtm.date.today(), t)
        )
        elapsed = clock_time - midnight
        return int(elapsed.total_seconds() / interval.total_seconds())

    def _find_pending_order(self, order_no) -> StockOrder:
        return find(lambda x: x.order_no == order_no, self.pending_orders)

    def _find_position(self, asset_code) -> StockPosition:
        return find(lambda x: x.asset_code == asset_code, self.positions)

    def _recalc_average_purchase_price(self, asset_code, quantity, price):
        p = self._find_position(asset_code)
        total_value = price * quantity
        total_quantity = quantity

        if p is not None:
            prev_value = Decimal(p.average_purchase_price * p.quantity)
            total_value += prev_value
            total_quantity += p.quantity

        return Decimal(total_value / total_quantity).quantize(
            Decimal("0.0001"), rounding=ROUND_HALF_UP
        )

    def _handle_order_event(
        self,
        event: OrderEvent,
    ):
        self.logger.debug(
            f"handle_order_event: accno={event.account_no} order_no={event.order_no} (org={event.org_order_no})\tside={event.side}\tcode={event.asset_code}\tfilled={event.filled_quantity}\torder_qty={event.quantity}\tevent_type={event.event_type}"
        )

        if event.account_no != self.account_no:
            return

        if event.event_type == "accepted":
            self._handle_accept_order_event(event)

        elif event.event_type == "cancelled":
            self._handle_cancel_order_event(event)

        elif event.event_type == "executed":
            self._handle_execution_order_event(event)

    def _handle_accept_order_event(self, event: OrderEvent):
        self.logger.debug(
            f"accept event: order_no={event.order_no} (org={event.org_order_no}) qty={event.quantity} side={event.side}"
        )
        order = StockOrder(
            order_no=event.order_no,
            asset_code=event.asset_code,
            side=event.side,
            order_type=event.order_type,
            quantity=event.quantity,
            price=event.price,
            filled_quantity=0,
            pending_quantity=event.quantity,
            order_time=event.filled_time,
            org_order_no=event.org_order_no,
        )
        self.pending_orders.append(order)

        org_order = self._find_pending_order(event.org_order_no)
        if org_order is not None:
            if org_order.pending_quantity == event.quantity:
                self.pending_orders.remove(org_order)
            else:
                org_order.pending_quantity -= event.quantity

            self._notify_pending_order_update("accepted", order)

        if event.side == OrderSide.SELL:
            position = self._find_position(event.asset_code)
            if position is not None:
                position.sell_possible_quantity -= event.quantity

    def _handle_cancel_order_event(self, event: OrderEvent):
        self.logger.debug(f"cancel event: order_no={event.order_no}")

        order = self._find_pending_order(event.order_no)
        if order is not None:
            self.pending_orders.remove(order)

            order.pending_quantity = 0
            order.is_pending = False

            if event.side == OrderSide.SELL:
                position = self._find_position(event.asset_code)
                if position is not None:
                    position.sell_possible_quantity += order.pending_quantity

            self._notify_pending_order_update("cancelled", order)

    def _handle_execution_order_event(self, event: OrderEvent):
        order_no = event.order_no
        side = event.side
        asset_code = event.asset_code
        total_filled_quantity = event.filled_quantity

        order = self._find_pending_order(event.order_no)
        position = self._find_position(event.asset_code)
        position_event_type = "modified"

        if order:
            order.filled_quantity += event.filled_quantity
            order.pending_quantity -= event.filled_quantity
            total_filled_quantity = order.filled_quantity

        self.logger.debug(
            f"execution event: order_no={event.order_no} filled={total_filled_quantity} total={event.quantity} filled_price={event.filled_price}"
        )

        if side == OrderSide.SELL:
            if position:
                position.quantity -= event.filled_quantity

                if position.quantity == 0:
                    self.positions.remove(position)
                    position_event_type = "removed"

            else:
                raise Exception("Position not found")

        elif side == OrderSide.BUY:
            if position:
                position.average_purchase_price = self._recalc_average_purchase_price(
                    asset_code, event.filled_quantity, event.filled_price
                )
                position.quantity += event.filled_quantity
                position.sell_possible_quantity += event.filled_quantity
            else:
                position = StockPosition(
                    asset_code=asset_code,
                    asset_name=self.tickers.get(asset_code, {}).get("name", ""),
                    quantity=event.filled_quantity,
                    sell_possible_quantity=event.filled_quantity,
                    average_purchase_price=Decimal(event.filled_price),
                )
                self.positions.append(position)
                position_event_type = "added"

        partial = event.quantity != total_filled_quantity

        if not partial and order:
            self.pending_orders.remove(order)

        self._save_trading_history(
            asset_code,
            side,
            order_no,
            event.filled_price,
            total_filled_quantity,
            position.average_purchase_price,
            event.filled_time,
            partial,
        )
        self._notify_pending_order_update("partial" if partial else "completed", order)
        self._notify_position_update(position_event_type, position)

    def _refresh_positions(self):
        self.positions = self.simple_api.get_positions()

    def _refresh_pending_orders(self):
        old_pending_orders = {}
        for o in self.pending_orders:
            old_pending_orders[o.order_no] = o

        self.pending_orders = self.simple_api.get_pending_orders()
        for o in self.pending_orders:
            old_pending_order = old_pending_orders.get(o.order_no)
            if old_pending_order is not None:
                o.average_purchase_price = old_pending_order.average_purchase_price

    def _notify_pending_order_update(self, status: str, order: StockOrder):
        if self.on_pending_order_update is not None:
            self.on_pending_order_update(status, order)

    def _notify_position_update(self, type: str, position: StockPosition = None):
        if self.on_position_update is not None:
            self.on_position_update(type, position)

    def _save_trading_history(
        self,
        asset_code: str,
        side: OrderSide,
        order_no: str,
        filled_price: int,
        filled_quantity: int,
        average_purchase_price: Decimal = None,
        executed_time: dtm.datetime = None,
        partial: bool = False,
    ):
        is_equity = self.tickers[asset_code]["type"] == "EQUITY"
        fee = filled_price * filled_quantity * self.fee_rate
        tax = 0
        pnl = None
        pnl_rate = None

        if side == OrderSide.SELL:
            sell_value = filled_price * filled_quantity
            if is_equity:
                tax = sell_value * self.tax_rate
            buy_value = average_purchase_price * filled_quantity
            buy_fee = buy_value * self.fee_rate
            pnl = sell_value - buy_value - fee - tax - buy_fee
            pnl_rate = pnl / buy_value * 100 if buy_value != 0 else 0

        data = TradingHistory(
            date=dtm.date.today().strftime("%Y%m%d"),
            order_no=order_no,
            side="buy" if side == OrderSide.BUY else "sell",
            asset_code=asset_code,
            quantity=filled_quantity,
            filled_price=filled_price,
            average_purchase_price=float(average_purchase_price),
            tax=float(tax) if tax is not None else None,
            fee=float(fee) if fee is not None else None,
            pnl=float(pnl) if pnl is not None else None,
            pnl_rate=float(pnl_rate) if pnl_rate is not None else None,
            executed_time=(
                int(executed_time.timestamp() * 1000) if executed_time else None
            ),
            partial=partial,
        )

        self._send_trading_history(data)

    def _send_trading_history(self, history: TradingHistory):
        url = f"{c.PYQQQ_API_URL}/analytics/trades/{history.order_no}"

        data = asdict(history)
        data["brokerage"] = (
            "ebest" if isinstance(self.simple_api, EBestSimpleDomesticStock) else "kis"
        )
        data["account_no"] = self.account_no

        strategy_name = os.getenv("STRATEGY_NAME")
        if strategy_name is not None:
            data["strategy_name"] = strategy_name

        positions = []
        for p in self.positions:
            d = asdict(p)
            d["average_purchase_price"] = float(d["average_purchase_price"])
            positions.append(d)

        r = send_request("POST", url, json=data)
        raise_for_status(r)

        self.logger.info(f"save trading history: {data}")

    def save_positions(self):
        strategy_name = os.getenv("STRATEGY_NAME")
        if strategy_name is None:
            return

        url = f"{c.PYQQQ_API_URL}/analytics/positions"

        positions = []

        for p in self.simple_api.get_positions():
            d = asdict(p)
            d["average_purchase_price"] = float(d["average_purchase_price"])
            d["current_pnl"] = float(d["current_pnl"]) if "current_pnl" in d else None
            positions.append(d)

        account = self.simple_api.get_account()
        account["pnl_rate"] = (
            float(account["pnl_rate"]) if "pnl_rate" in account else None
        )

        req_body = {
            "date": dtm.date.today().strftime("%Y%m%d"),
            "brokerage": (
                "ebest"
                if isinstance(self.simple_api, EBestSimpleDomesticStock)
                else "kis"
            ),
            "account_no": self.account_no,
            "positions": positions,
            "account": account,
            "strategy_name": strategy_name,
        }

        print(req_body)

        r = send_request("POST", url, json=req_body)
        raise_for_status(r)

        self.logger.info(f"save positions: {positions}")
